/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc.hsql;

import org.wso2.registry.RegistryException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import javax.sql.DataSource;
import java.sql.*;

/**
 * Creates the tables required for HSQL database.
 */
public class DatabaseCreator {

    public static final Log log = LogFactory.getLog(DatabaseCreator.class);

    //////////////////////////////////////////////////////////////////////////////////
    // Tables related to current versions of resources and metadata                 //
    //////////////////////////////////////////////////////////////////////////////////

    private static final String resourceTable =
            "CREATE TABLE RESOURCE (" +
            "RID VARCHAR (50)," +
            "PATH VARCHAR (500) NOT NULL," +
            "MEDIA_TYPE VARCHAR (500)," +
            "COLLECTION INTEGER NOT NULL," +
            "CREATOR VARCHAR (500)," +
            "CREATED_TIME TIMESTAMP," +
            "LAST_UPDATOR VARCHAR (500)," +
            "LAST_UPDATED_TIME TIMESTAMP," +
            "DESCRIPTION VARCHAR (500)," +
            "CONTENT_ID VARCHAR (50)," +
            "PRIMARY KEY (RID)," +
            "FOREIGN KEY (CONTENT_ID) REFERENCES CONTENT (CONTENT_ID)," +
            "UNIQUE(PATH));";

    private static final String contentTable =
            "CREATE TABLE CONTENT (" +
            "CONTENT_ID VARCHAR (50)," +
            "CONTENT_DATA BINARY," +
            "PRIMARY KEY (CONTENT_ID));";

    // Table to store parent child relationships in current versions of resources. Collection
    // which is referred in PARENT_RID is not allowed to delete until the relationship is
    // explicitly removed from this table. This is to ensure that collections are not deleted
    // till all its children are removed. If a child resource (which has no children) is deleted
    // its parent child relationship is automatically deleted.

    private static final String dependencyTable =
            "CREATE TABLE DEPENDENCY (" +
            "DEPENDENCY_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
            "PARENT_RID VARCHAR (50) NOT NULL," +
            "CHILD_RID VARCHAR (50) NOT NULL," +
            "PRIMARY KEY (DEPENDENCY_ID)," +
            "UNIQUE (PARENT_RID, CHILD_RID)," +
            "FOREIGN KEY (PARENT_RID) REFERENCES RESOURCE (RID) ON DELETE CASCADE," +
            "FOREIGN KEY (CHILD_RID) REFERENCES RESOURCE (RID) ON DELETE CASCADE);";

    private static final String propertyTable =
            "CREATE TABLE PROPERTY (" +
            "PROPERTY_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
            "RID VARCHAR (50) NOT NULL," +
            "NAME VARCHAR (100) NOT NULL," +
            "VALUE VARCHAR (500) NOT NULL," +
            "PRIMARY KEY (PROPERTY_ID)," +
            "UNIQUE (RID, NAME, VALUE)," +
            "FOREIGN KEY (RID) REFERENCES RESOURCE (RID) ON DELETE CASCADE);";

    private static final String associationTable =
            "CREATE TABLE ASSOCIATION (" +
            "ASSOCIATION_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
            "RID1 VARCHAR (50) NOT NULL," +
            "RID2 VARCHAR (50) NOT NULL," +
            "ASSOCIATION_TYPE VARCHAR (20) NOT NULL," +
            "UNIQUE (RID1, RID2, ASSOCIATION_TYPE)," +
            "FOREIGN KEY (RID1) REFERENCES RESOURCE (RID) ON DELETE CASCADE," +
            "FOREIGN KEY (RID2) REFERENCES RESOURCE (RID) ON DELETE CASCADE);";

    // Note that comments, tags and ratings are not versioned. Therefore, when a resource is deleted
    // its comments, etc should not be deleted. When a deleted resource is restored, its
    // comments, etc are reassociated with it.

    private static final String tagTable =
            "CREATE TABLE TAG (" +
            "TAG_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
            "TAG_NAME VARCHAR (500) NOT NULL," +
            "RID VARCHAR (50) NOT NULL," +
            "USER_ID VARCHAR (20) NOT NULL," +
            "TAGGED_TIME TIMESTAMP NOT NULL," +
            "PRIMARY KEY (TAG_ID)," +
            "UNIQUE (TAG_NAME, RID, USER_ID));";

    private static final String commentTable =
            "CREATE TABLE COMMENT (" +
            "COMMENT_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
            "RID VARCHAR (50) NOT NULL," +
            "USER_ID VARCHAR (20) NOT NULL," +
            "COMMENT_TEXT VARCHAR (500) NOT NULL," +
            "COMMENTED_TIME TIMESTAMP NOT NULL," +
            "PRIMARY KEY (COMMENT_ID));";

    private static final String ratingTable =
            "CREATE TABLE RATING (" +
            "RATING_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
            "RID VARCHAR (50) NOT NULL," +
            "USER_ID VARCHAR (20) NOT NULL," +
            "RATING INTEGER NOT NULL," +
            "RATED_TIME TIMESTAMP NOT NULL," +
            "PRIMARY KEY (RATING_ID));";

    private static final String logTable =
            "CREATE TABLE LOG (" +
            "LOG_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
            "PATH VARCHAR (50)," +
            "USER_ID VARCHAR (20) NOT NULL," +
            "LOGGED_TIME TIMESTAMP NOT NULL," +
            "ACTION INTEGER NOT NULL," +
            "ACTION_DATA VARCHAR (500)," +
            "PRIMARY KEY (LOG_ID));";

    //////////////////////////////////////////////////////////////////////////////////
    // Tables related to old versions of resources and metadata                     //
    //////////////////////////////////////////////////////////////////////////////////

    private static final String resourceVersionTable =
            "CREATE TABLE RESOURCE_VERSION (" +
            "RID VARCHAR (50)," +
            "PATH VARCHAR (500) NOT NULL," +
            "MEDIA_TYPE VARCHAR (500)," +
            "DIRECTORY INTEGER NOT NULL," +
            "CREATOR VARCHAR (500)," +
            "CREATED_TIME TIMESTAMP," +
            "LAST_UPDATOR VARCHAR (500)," +
            "LAST_UPDATED_TIME TIMESTAMP," +
            "DESCRIPTION VARCHAR (500)," +
            "CONTENT_ID VARCHAR (50)," +
            "FOREIGN KEY (CONTENT_ID) REFERENCES CONTENT_VERSION (CONTENT_VERSION_ID)," +
            "PRIMARY KEY (RID)," +
            "UNIQUE(PATH));";

    private static final String dependencyVersionTable =
            "CREATE TABLE DEPENDENCY_VERSION (" +
            "DEPENDENCY_VERSION_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
            "PARENT_RID VARCHAR (50) NOT NULL," +
            "PARENT_VERSION INTEGER NOT NULL," +
            "CHILD_RID VARCHAR (50) NOT NULL," +
            "PRIMARY KEY (DEPENDENCY_VERSION_ID)," +
            "UNIQUE (PARENT_RID, PARENT_VERSION, CHILD_RID)," +
            "FOREIGN KEY (PARENT_RID) REFERENCES RESOURCE_VERSION (RID)," +
            "FOREIGN KEY (CHILD_RID) REFERENCES RESOURCE_VERSION (RID));";

    private static final String contentVersionTable =
            "CREATE TABLE CONTENT_VERSION (" +
            "CONTENT_VERSION_ID VARCHAR (50)," +
            "CONTENT_DATA BINARY," +
            "PRIMARY KEY (CONTENT_VERSION_ID));";

    private static final String propertyVersionTable =
            "CREATE TABLE PROPERTY_VERSION (" +
            "PROPERTY_VERSION_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
            "RID VARCHAR (50) NOT NULL," +
            "VERSION INTEGER NOT NULL," +
            "NAME VARCHAR (100) NOT NULL," +
            "VALUE VARCHAR (500) NOT NULL," +
            "PRIMARY KEY (PROPERTY_VERSION_ID)," +
            "UNIQUE (RID, VERSION, NAME, VALUE)," +
            "FOREIGN KEY (RID) REFERENCES RESOURCE_VERSION (RID));";

    private static final String associationVersionTable =
            "CREATE TABLE ASSOCIATION_VERSION (" +
            "ASSOCIATION_VERSION_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
            "RID1 VARCHAR (50) NOT NULL," +
            "RID1_VERSION INTEGER NOT NULL," +
            "RID2 VARCHAR (50) NOT NULL," +
            "RID2_VERSION INTEGER NOT NULL," +
            "ASSOCIATION_TYPE VARCHAR (20)," +
            "PRIMARY KEY (ASSOCIATION_VERSION_ID)," +
            "UNIQUE (RID1, RID1_VERSION, RID2, RID2_VERSION, ASSOCIATION_TYPE)," +
            "FOREIGN KEY (RID1) REFERENCES RESOURCE_VERSION (RID)," +
            "FOREIGN KEY (RID2) REFERENCES RESOURCE_VERSION (RID));";

    //////////////////////////////////////////////////////////////////////////////////
    // Tables related to snapshot management                                        //
    //////////////////////////////////////////////////////////////////////////////////

    // UUID is used for snapshot ID, so that snapshots can be achieved and reloaded without
    // colliding with existing snapshots.

    private static final String snapshotTable =
            "CREATE TABLE SNAPSHOT (" +
            "SNAPSHOT_ID VARCHAR (50)," +
            "ROOT_ID VARCHAR (50) NOT NULL," +
            "ROOT_SNAPSHOT_NUMBER INTEGER NOT NULL," +
            "PRIMARY KEY (SNAPSHOT_ID)," +
            "UNIQUE (ROOT_ID, ROOT_SNAPSHOT_NUMBER));";

    private static final String snapshotResourceVersionTable =
            "CREATE TABLE SNAPSHOT_RESOURCE_VERSION (" +
            "SRV_ID INTEGER GENERATED BY DEFAULT AS IDENTITY," +
            "SNAPSHOT_ID VARCHAR (50) NOT NULL," +
            "RID VARCHAR (50) NOT NULL," +
            "VERSION INTEGER NOT NULL," +
            "UNIQUE (SNAPSHOT_ID, RID, VERSION));";


    public static synchronized void createDatabase(DataSource dataSource) throws RegistryException {

        Connection conn = null;

        try {
            conn = dataSource.getConnection();
            conn.setAutoCommit(false);

            DatabaseMetaData metaData = conn.getMetaData();
            ResultSet result = metaData.getTables(null, null, "RESOURCE", null);
            if (result.next()) {
                return;
            }

            ResultSet result2 = metaData.getTables(null, null, "resource", null);
            if (result2.next()) {
                return;
            }

            Statement s = conn.createStatement();

            // create current version related tables
            s.executeUpdate(contentTable);
            s.executeUpdate(resourceTable);
            s.executeUpdate(dependencyTable);
            s.executeUpdate(propertyTable);
            s.executeUpdate(associationTable);
            s.executeUpdate(tagTable);
            s.executeUpdate(commentTable);
            s.executeUpdate(ratingTable);
            s.executeUpdate(logTable);

            // create old versions related tables
            s.executeUpdate(contentVersionTable);
            s.executeUpdate(resourceVersionTable);
            s.executeUpdate(dependencyVersionTable);
            s.executeUpdate(propertyVersionTable);
            s.executeUpdate(associationVersionTable);

            // create snapshot management tables
            s.executeUpdate(snapshotTable);
            s.executeUpdate(snapshotResourceVersionTable);

            conn.commit();

        } catch (SQLException e) {

            String msg = "Could not create the database tables of WSO2 Registry. " + e.getMessage();
            log.fatal(msg, e);

            try {
                conn.rollback();
            } catch (SQLException e1) {
                String msg2 = "Failed to rollback the unsuccessful creation of Registry " +
                        "database tables. " + e1.getMessage();
                log.error(msg2, e1);
            }

            throw new RegistryException(msg, e);

        } finally {

            if (conn != null) {
                try {
                    conn.close();
                } catch (SQLException e) {
                    String msg3 = "Failed to close the database connection used for creating " +
                            "Registry database tables. " + e.getMessage();
                    log.error(msg3, e);
                }
            }
        }
    }
}
